#include <iostream>
#include <set>
#include <string>
#include <sstream>
#include <vector>
using std::cout;
using std::endl;
using std::set;
using std::string;
using std::ostringstream;
using std::vector;

struct piece
{
	bool ulc;
	bool us;
	bool urc;
	bool rs;
	bool drc;
	bool ds;
	bool dlc;
	bool ls;
};

vector<piece> pieces= {
	{false, false, false, false, false, false, false, false},
	{false,  true, false, false, false, false, false, false},
	{false, false, false,  true, false, false, false, false},
	{false,  true, false,  true, false, false, false, false},
	{false,  true,  true,  true, false, false, false, false},
	{false, false, false, false, false,  true, false, false},
	{false,  true, false, false, false,  true, false, false},
	{false, false, false,  true, false,  true, false, false},
	{false,  true, false,  true, false,  true, false, false},
	{false,  true,  true,  true, false,  true, false, false},
	{false, false, false,  true,  true,  true, false, false},
	{false,  true, false,  true,  true,  true, false, false},
	{false,  true,  true,  true,  true,  true, false, false},
	{false, false, false, false, false, false, false,  true},
	{false,  true, false, false, false, false, false,  true},
	{ true,  true, false, false, false, false, false,  true},
	{false, false, false,  true, false, false, false,  true},
	{false,  true, false,  true, false, false, false,  true},
	{ true,  true, false,  true, false, false, false,  true},
	{false,  true,  true,  true, false, false, false,  true},
	{ true,  true,  true,  true, false, false, false,  true},
	{false, false, false, false, false,  true, false,  true},
	{false,  true, false, false, false,  true, false,  true},
	{ true,  true, false, false, false,  true, false,  true},
	{false, false, false,  true, false,  true, false,  true},
	{false,  true, false,  true, false,  true, false,  true},
	{ true,  true, false,  true, false,  true, false,  true},
	{false,  true,  true,  true, false,  true, false,  true},
	{ true,  true,  true,  true, false,  true, false,  true},
	{false, false, false,  true,  true,  true, false,  true},
	{false,  true, false,  true,  true,  true, false,  true},
	{ true,  true, false,  true,  true,  true, false,  true},
	{false,  true,  true,  true,  true,  true, false,  true},
	{ true,  true,  true,  true,  true,  true, false,  true},
	{false, false, false, false, false,  true,  true,  true},
	{false,  true, false, false, false,  true,  true,  true},
	{ true,  true, false, false, false,  true,  true,  true},
	{false, false, false,  true, false,  true,  true,  true},
	{false,  true, false,  true, false,  true,  true,  true},
	{ true,  true, false,  true, false,  true,  true,  true},
	{false,  true,  true,  true, false,  true,  true,  true},
	{ true,  true,  true,  true, false,  true,  true,  true},
	{false, false, false,  true,  true,  true,  true,  true},
	{false,  true, false,  true,  true,  true,  true,  true},
	{ true,  true, false,  true,  true,  true,  true,  true},
	{false,  true,  true,  true,  true,  true,  true,  true},
	{ true,  true,  true,  true,  true,  true,  true,  true},
	{false, false, false, false, false, false, false, false}
	// The piece with all sides and corners closed is duplicated because there
	// is one unset piece in the solution.
};
piece closed_piece= {false, false, false, false, false, false, false, false};

size_t unset_index= 256;
vector<vector<size_t> > board;
set<size_t> unused_pieces;
size_t x_max= 7;
size_t y_max= 5;

void init_board()
{
	board.resize(x_max+1);
	for(auto&& column : board)
	{
		column.resize(y_max+1);
		for(auto&& part : column)
		{
			part= unset_index;
		}
	}
	unused_pieces.clear();
	for(size_t i= 0; i < pieces.size(); ++i)
	{
		unused_pieces.insert(i);
	}
}

void add_piece_id_to_usage_list(size_t id, set<size_t>& usage_list)
{
	usage_list.insert(id);
}

void remove_piece_id_from_usage_list(size_t id, set<size_t>& usage_list)
{
	auto iter= usage_list.find(id);
	if(iter != usage_list.end())
	{
		usage_list.erase(iter);
	}
}

bool left_right_match(piece a, piece b)
{
	return a.ulc == b.urc && a.ls == b.rs && a.dlc == b.drc;
}

bool up_down_match(piece a, piece b)
{
	return a.ulc == b.dlc && a.us == b.ds && a.urc == b.drc;
}

bool piece_fits(piece curr, size_t x, size_t y)
{
	if(x > 0)
	{
		if(board[x-1][y] != unset_index)
		{
			piece leftward= pieces[board[x-1][y]];
			if(!left_right_match(curr, leftward))
			{
				return false;
			}
		}
		if(x >= x_max && !left_right_match(closed_piece, curr))
		{
			return false;
		}
	}
	else if(!left_right_match(curr, closed_piece))
	{
		return false;
	}
	if(y > 0)
	{
		if(board[x][y-1] != unset_index)
		{
			piece upward= pieces[board[x][y-1]];
			if(!up_down_match(curr, upward))
			{
				return false;
			}
		}
		if(y >= y_max && !up_down_match(closed_piece, curr))
		{
			return false;
		}
	}
	else if(!up_down_match(curr, closed_piece))
	{
		return false;
	}
	// Don't bother comparing to the positive direction pieces because the
	// algorithm guarantees they are unset.
	return true;

	if(x < x_max)
	{
		if(board[x+1][y] != unset_index)
		{
			piece rightward= pieces[board[x+1][y]];
			if(!left_right_match(rightward, curr))
			{
				return false;
			}
		}
	}
	if(y < y_max)
	{
		if(board[x][y+1] != unset_index)
		{
			piece downward= pieces[board[x][y+1]];
			if(!up_down_match(downward, curr))
			{
				return false;
			}
		}
	}
	return true;
}

size_t find_solution_call_count= 0;
size_t freq= 100000000;
size_t dump_count= 0;
size_t full_dump_count= 0;
void dump_board()
{
	cout << dump_count << " bored yet?" << endl;
	for(size_t y= 0; y <= y_max; ++y)
	{
		for(size_t x= 0; x <= x_max; ++x)
		{
			cout << board[x][y] << ' ';
		}
		cout << endl;
	}
}

int bbit(bool b) { return b ? 1 : 0; }

void full_dump_row(size_t y)
{
	ostringstream top;
	ostringstream mid;
	ostringstream bot;
	for(size_t x= 0; x <= x_max; ++x)
	{
		size_t piend= board[x][y];
		if(piend == unset_index)
		{
			top << "   ";
			mid << "   ";
			bot << "   ";
		}
		else
		{
			piece pie= pieces[piend];
			top << bbit(pie.ulc) << bbit(pie.us) << bbit(pie.urc);
			mid << bbit(pie.ls) << ' ' << bbit(pie.rs);
			bot << bbit(pie.dlc) << bbit(pie.ds) << bbit(pie.drc);
		}
	}
	cout << top.str() << endl;
	cout << mid.str() << endl;
	cout << bot.str() << endl;
}

void full_dump_board()
{
	cout << "full dump: " << full_dump_count << endl;
	for(size_t y= 0; y <= y_max; ++y)
	{
		full_dump_row(y);
	}
	++full_dump_count;
}

void dumper()
{
	if(find_solution_call_count % freq == 0)
	{
		if(dump_count % 100 == 0)
		{
			full_dump_board();
		}
		else
		{
			dump_board();
		}
		++dump_count;
	}
	++find_solution_call_count;
}

bool find_solution(size_t x, size_t y)
{
	auto try_piece= unused_pieces.begin();
	while(try_piece != unused_pieces.end())
	{
		dumper();
		size_t use_piece= *try_piece;
		if(!piece_fits(pieces[use_piece], x, y))
		{
			++try_piece;
			continue;
		}
		board[x][y]= use_piece;
		remove_piece_id_from_usage_list(use_piece, unused_pieces);
		size_t next_x= x+1;
		size_t next_y= y;
		if(next_x > x_max)
		{
			next_x= 0;
			next_y+= 1;
		}
		if(next_y > y_max)
		{
			return true;
		}
		bool sub_suck= find_solution(next_x, next_y);
		if(sub_suck)
		{
			return true;
		}
		add_piece_id_to_usage_list(use_piece, unused_pieces);
		++try_piece;
	}
	return false;
}

int main()
{
	init_board();
	size_t curr_x= 0;
	size_t curr_y= 0;
	bool solution_exists= find_solution(curr_x, curr_y);
	full_dump_board();
	cout << "Solution exists? " << solution_exists << endl;
	return 0;
}
